#ifndef ARCHIVE_H
#define ARCHIVE_H


#include "refcount.h"
#include <stdlib.h>
#include <stdio.h>


class ArchiveDataSource
{
	public:
	virtual ~ArchiveDataSource(void) {}
	virtual bool IsValid(void) {return false;}
	virtual void Read(void *Buff,int Len)=0;
	virtual void Write(const void *Buff,int Len)=0;
	virtual int GetPos(void)=0;
};

class FILEArchiveDataSource:public ArchiveDataSource
{
	FILE *f;
	public:
	
	FILEArchiveDataSource(const char *FName,bool Read) {f=fopen(FName,Read?"rb":"wb");}
	~FILEArchiveDataSource(void) {if (f!=NULL) fclose(f);}
	virtual bool IsValid(void) {return f!=NULL;}
	virtual void Read(void *Buff,int Len) {fread(Buff,1,Len,f);}
	virtual void Write(const void *Buff,int Len) {fwrite(Buff,1,Len,f);}
	virtual int GetPos(void) {return ftell(f);}
};



#define ARCHIVABLE(X)\
     virtual void DoOperation(Archive &archiv,int Version);\
     virtual const char *GetClassName(void) {return #X;}\
     virtual Archivable *ConstructClassInstance(void) {return new X;}\
     virtual void DestructClassInstance(Archivable *I) {if (I) delete (X *)I;}\
     virtual bool IsSupportedForLoad(int Version);\
     virtual int GetSavedVersion(void);\
     virtual void Invalidate(void);


#define ARCHIVABLEEX(X,D,VLoadMin,VSave,VLoadMax)\
     virtual void DoOperation(Archive &archiv,int Version);\
     virtual const char *GetClassName(void) {return #X #D;}\
     virtual Archivable *ConstructClassInstance(void) {return new X;}\
     virtual void DestructClassInstance(Archivable *I) {if (I) delete (X *)I;}\
     virtual bool IsSupportedForLoad(int Version) {return VLoadMin<=Version && Version<=VLoadMax;}\
     virtual int GetSavedVersion(void) {return VSave;}\
     virtual void Invalidate(void);

#define ARCHIVABLEEXDIRECTIMPL(X,D,VLoadMin,VSave,VLoadMax)\
     virtual const char *GetClassName(void) {return #X #D;}\
     virtual Archivable *ConstructClassInstance(void) {return new X;}\
     virtual void DestructClassInstance(Archivable *I) {if (I) delete (X *)I;}\
     virtual bool IsSupportedForLoad(int Version) {return VLoadMin<=Version && Version<=VLoadMax;}\
     virtual int GetSavedVersion(void) {return VSave;}

#define ARCHIVABLENOINVAL(X,D,VLoadMin,VSave,VLoadMax)\
     virtual void DoOperation(Archive &archiv,int Version);\
     virtual const char *GetClassName(void) {return #X #D;}\
     virtual Archivable *ConstructClassInstance(void) {return new X;}\
     virtual void DestructClassInstance(Archivable *I) {if (I) delete (X *)I;}\
     virtual bool IsSupportedForLoad(int Version) {return VLoadMin<=Version && Version<=VLoadMax;}\
     virtual int GetSavedVersion(void) {return VSave;}\
     virtual void Invalidate(void) {}

#define ARCHIVABLENOIMPL(X)\
	 virtual void DoOperation(Archive &archiv,int Version) {assert(!"Unsupported");} \
     virtual const char *GetClassName(void) {return #X;}\
     virtual Archivable *ConstructClassInstance(void) {return new X;}\
     virtual void DestructClassInstance(Archivable *I) {if (I) delete (X *)I;}\
     virtual bool IsSupportedForLoad(int Version) {return false;}\
     virtual int GetSavedVersion(void) {return 0;}\
     virtual void Invalidate(void) {}


#define STOREENUM(Arch,Value,EnumType)\
     {\
     	int TmpVal=(int)Value;\
     	Arch<<=TmpVal;\
		Value=(EnumType)TmpVal;\
     }



class Archive;
class Archivable
{
     public:
     virtual ~Archivable(void) {}

     virtual Archivable *ConstructClassInstance(void)=0; //vytvori instanci classy
     virtual void DestructClassInstance(Archivable *)=0; //zrusi instanci classy
     virtual void DoOperation(Archive &archiv,int Version)=0; 
	//provede operaci nad classou
	//verse je pri ukladani rovna hodnote z GetSavedVersion()
	//pri nacteni je to hodnota verse ze souboru
     virtual const char *GetClassName(void)=0; //retezec se jmenem tridy     
     virtual bool IsSupportedForLoad(int Version)=0; //vraci true pokud je pro nacteni podporovana dana verse
     virtual int GetSavedVersion(void)=0; //vraci cislo verse ktera je ukladana
     virtual void Invalidate(void)=0; //volano pokud selze nacteni 
     
};


class Archive:public RefCounter
				//abstraktni predek pro archivy, slouzi pro zakryti 
				//implementace archivu aby se zabranilo kolizim pri mezi STL a MFC
{
	 protected:
	 ArchiveDataSource *DataSource;
     	 enum OperationModeType {UNKNOWN,READ,WRITE,CALCULATESIZE};
     	 public:
     	 OperationModeType OperationMode; //jaka operace zrovna probiha

	 Archive(void) {DataSource=NULL;OperationMode=UNKNOWN;}
		 
	 virtual ~Archive(void) {DoneArchive();}
	 
	 virtual void RegisterClass(Archivable *ClassFactory); //volat pred initarchive
	 virtual void InitArchive(ArchiveDataSource *DS,bool ForRead); //volat po vytvoreni archivu
	 virtual void DoneArchive(void);
     
	 //zjistovani stavu
     bool DoLoad(void) {return OperationMode==READ;}
     bool DoSave(void) {return (OperationMode==WRITE) || (OperationMode==CALCULATESIZE);}
     bool DoCalc(void) {return OperationMode==CALCULATESIZE;}

	 //zakladni operace
	 Archive &operator<<=(bool &Value);
	 Archive &operator<<=(int &Value);
	 Archive &operator<<=(unsigned long &Value);
     Archive &operator<<=(float &Value);
	 Archive &operator<<=(long double &Value);
     Archive &operator<<=(char &Value);
	 virtual Archive &operator<<=(Archivable &Value)=0;
	 virtual Archive &operator<<=(Archivable *&Value)=0;
	 
	 //rozsirene operace
     void Raw(void *Buff,int Size); 
     void Raw(const void *Buff,int Size); 

	
	 static Archive *CreateArchive(void); //vytvori instanci skryteho archivu

	 protected:
	 int GetOffset(void);
};

template<class X>
Archive &operator<<=(Archive &arch,RefCounterHolder<X,true> &s)
{
	Archivable *Tmp=s.GetValue();
	arch<<=Tmp;
	s.SetValue((X *)Tmp);
	return arch;
}

template<class X>
Archive &operator<<=(Archive &arch,RefCounterHolder<X,false> &s)
{
	Archivable *Tmp=s.GetValue();
	arch<<=Tmp;
	s.SetValue((X *)Tmp);
	return arch;
}


#endif